"""Common module for y and z stitcher"""

from __future__ import annotations

import os
import argparse
import logging
import signal
import sys
import functools

from tqdm import tqdm

import silx
from silx.gui import qt
from silx.io.url import DataUrl

from nabu.pipeline.config import generate_nabu_configfile
from nabu.stitching.config import get_default_stitching_config, SECTIONS_COMMENTS
from nabu.stitching.config import StitchingType

from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.settings import SlurmSettingsMode
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.core.process.stitching.nabustitcher import StitcherTask
from tomwer.io.utils.tomoobj import get_tomo_objs_instances
from tomwer.gui.utils.splashscreen import getMainSplashScreen
from tomwer.gui.stitching.StitchingWindow import ZStitchingWindow as _ZStitchingWindow
from tomwer.gui.stitching.StitchingWindow import YStitchingWindow as _YStitchingWindow
from tomwer.gui.stitching import action as stitching_action
from tomwer.gui.cluster.slurm import SlurmSettingsWidget
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume import HDF5Volume
from tomwer.core.volume import MultiTIFFVolume
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.tomwer_object import TomwerObject


class MainWidget(qt.QTabWidget):
    DEFAULT_NB_WORKERS = 10

    def __init__(
        self,
        axis: int,
        parent=None,
    ) -> None:
        super().__init__(parent)
        if axis == 0:
            self._stitchingConfigWindow = _ZStitchingWindow(
                parent=self,
                with_configuration_action=False,
            )
        elif axis == 1:
            self._stitchingConfigWindow = _YStitchingWindow(
                parent=self, with_configuration_action=False
            )
        else:
            raise ValueError("stitching is only available along axis 0 and 1")

        self.addTab(self._stitchingConfigWindow, "stitching config")
        settingsClass = SlurmSettingsMode.get_settings_class(
            SlurmSettingsMode.STITCHING
        )
        self._slurmConfig = SlurmSettingsWidget(
            parent=self, n_gpu=settingsClass.N_GPUS_PER_WORKER, jobLimitation=None
        )
        self._slurmConfig.setConfiguration(
            {
                "cpu-per-task": settingsClass.N_CORES_PER_TASK,
                "n_tasks": settingsClass.N_TASKS,
                "n_jobs": settingsClass.N_JOBS,
                "memory": settingsClass.MEMORY_PER_WORKER,
                "partition": settingsClass.PARTITION,
                "n_gpus": settingsClass.N_GPUS_PER_WORKER,
                "job_name": settingsClass.PROJECT_NAME,
                "python_venv": settingsClass.PYTHON_VENV,
            }
        )

        style = qt.QApplication.style()
        help_icon = style.standardIcon(qt.QStyle.SP_TitleBarContextHelpButton)

        self._helpButton = qt.QPushButton(help_icon, "help")
        self._slurmConfig.layout().addWidget(
            self._helpButton, 99, 0, qt.Qt.AlignLeft, qt.Qt.AlignTop
        )

        self._stitchingConfigWindow.setCallbackToGetSlurmConfig(
            self._slurmConfig.getConfiguration
        )
        self._stitchingConfigWindow.setCallbackToSetSlurmConfig(
            self._slurmConfig.setConfiguration
        )
        self.addTab(self._slurmConfig, "slurm config")
        # add an option to clearly activate / deactivate slurm config
        slurm_config_idx = self.indexOf(self._slurmConfig)
        self._slurmCB = qt.QCheckBox(self)
        self.tabBar().setTabButton(
            slurm_config_idx,
            qt.QTabBar.LeftSide,
            self._slurmCB,
        )
        self._slurmCB.setChecked(True)

        # set up
        self._slurmConfig.setNWorkers(self.DEFAULT_NB_WORKERS)

        # connect signal / slot
        self._helpButton.released.connect(self._showSlurmSettingHelpForStitching)

    def close(self):
        self._stitchingConfigWindow.close()
        # requested for the waiting plot update
        super().close()

    def keyPressEvent(self, event):
        """
        To shortcut orange and make sure the `F5` <=> refresh stitching preview
        """
        modifiers = event.modifiers()
        key = event.key()

        if key == qt.Qt.Key_F5:
            self._stitchingConfigWindow._trigger_update_preview()
        elif key == qt.Qt.Key_O and modifiers == qt.Qt.KeyboardModifier.ControlModifier:
            self._stitchingConfigWindow._loadSettings()
        elif key == qt.Qt.Key_S and modifiers == qt.Qt.KeyboardModifier.ControlModifier:
            self._stitchingConfigWindow._saveSettings()

    def getStitchingConfiguration(self) -> dict:
        return self._stitchingConfigWindow.getConfiguration()

    def getClusterConfig(self) -> dict:
        return self._slurmConfig.getConfiguration()

    def _showSlurmSettingHelpForStitching(self):
        msg = qt.QMessageBox()
        msg.setWindowTitle("Help")
        msg.setText(
            """
            <b>Stitching Process Hints:</b>
            <ul>
            <li>
                <b>Processing Mode:</b> Stitching is performed using <b>CPU only</b> (GPU acceleration is not supported).
            </li>
            <li>
                <b>Memory Management:</b>
                <ul>
                <li>Stitching is processed in <b>chunks</b>, based on available memory.</li>
                <li>At least <b>one slice from each element</b> must be loaded into memory at all times.</li>
                </ul>
            </li>
            <li>
                <b>Job Parallelization:</b>
                <ul>
                <li>The stitching task is split into <b><code>n</code> parallel jobs</b> (where <code>n</code> = number of jobs).</li>
                <li>For optimal performance, we recommend each job processes a "<b>reasonable</b>" data size (e.g., <b>30 GB to 100 GB</b>).</li>
                </ul>
            </li>
            </ul>
            See <a href='https://tomwer.readthedocs.io/en/stable/applications/z_stitching.html'> application documentation</a> and <a href='https://tomwer.readthedocs.io/en/stable/tutorials/notebooks/standalone_apps/z_stitching.html'> stitching tutorial</a> for more details.
            """
        )
        msg.setIcon(qt.QMessageBox.Information)
        msg.exec()

    # expose API
    def addTomoObj(self, *args, **kwargs):
        self._stitchingConfigWindow.addTomoObj(*args, **kwargs)

    def setStitchingType(self, stitching_type):
        self._stitchingConfigWindow.setStitchingType(stitching_type)

    def getStitchingType(self):
        return self._stitchingConfigWindow.getStitchingType()

    def loadSettings(self, config_file):
        self._stitchingConfigWindow._loadSettings(config_file)

    def get_available_pre_processing_stitching_mode(self):
        return self._stitchingConfigWindow.get_available_pre_processing_stitching_mode()

    def get_available_post_processing_stitching_mode(self):
        return (
            self._stitchingConfigWindow.get_available_post_processing_stitching_mode()
        )

    def getTomoObjs(self):
        return self._stitchingConfigWindow.getTomoObjs()


class _StitcherThread(qt.QThread):
    DEFAULT_OUTPUT_CONFIG_NAME = "stitching_conf_autosave.conf"

    def __init__(self, parent, stitching_config, cluster_config) -> None:
        super().__init__(parent)
        assert isinstance(stitching_config, dict)
        assert isinstance(cluster_config, dict)
        self._stitching_config = stitching_config
        self._cluster_config = cluster_config

    def save_configuration(self):
        stitching_type = self._stitching_config["stitching"]["type"]
        if "preproc" in stitching_type:
            output_file = self._stitching_config.get("preproc", {}).get("location", "")
            output_folder = os.path.dirname(output_file)
        else:
            output_volume = self._stitching_config.get("postproc", {}).get(
                "output_volume", None
            )
            if output_volume is None:
                output_folder = ""
            else:
                volume = VolumeFactory.create_tomo_object_from_identifier(output_volume)
                output_folder = os.path.dirname(volume.data_url.file_path())

        os.makedirs(output_folder, exist_ok=True)
        generate_nabu_configfile(
            fname=os.path.join(
                output_folder,
                _StitcherThread.DEFAULT_OUTPUT_CONFIG_NAME,
            ),
            default_config=get_default_stitching_config(
                self._stitching_config["stitching"]["type"]
            ),
            comments=True,
            sections_comments=SECTIONS_COMMENTS,
            options_level="advanced",
            prefilled_values=self._stitching_config,
        )

    def run(self):
        self.save_configuration()

        task = StitcherTask(
            inputs={
                "stitching_config": self._stitching_config,
                "cluster_config": self._cluster_config,
                "progress": tqdm(desc="stitching"),
                "serialize_output_data": False,
            },
        )

        task.run()
        print("stitching finished")


class MainWindow(qt.QDialog):

    sigTomoObjAdded = qt.Signal(TomwerObject)

    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QVBoxLayout())
        # menu
        self._menuBar = qt.QMenuBar(self)
        self.layout().addWidget(self._menuBar)
        self._menu = qt.QMenu("&Stitching")
        self._menuBar.addMenu(self._menu)
        # save / load configuration
        self._loadConfigurationAction = stitching_action.LoadConfigurationAction(
            self, "&Load configuration"
        )
        self._menu.addAction(self._loadConfigurationAction)
        self._saveConfigurationAction = stitching_action.SaveConfigurationAction(
            self, "&Save configuration"
        )
        self._menu.addAction(self._saveConfigurationAction)
        # separator
        self._menu.addSeparator()
        # add tomo object
        self._addTomoObjectAction = stitching_action.AddTomoObjectAction(
            self, "&Add tomo object (volume or NXtomo)"
        )
        self._menu.addAction(self._addTomoObjectAction)

        # main window
        self.setWindowTitle("tomwer z-stitching")
        self._mainWindow = MainWidget(parent=self, axis=axis)
        self.layout().addWidget(self._mainWindow)

        # button
        types = qt.QDialogButtonBox.Apply
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)
        self._buttons.button(qt.QDialogButtonBox.Apply).setText("Launch stitching")

        # set up
        self._mainWindow._stitchingConfigWindow.setPreProcessingOutput(
            NXtomoScan(
                scan=os.path.join(os.getcwd(), "stitching", "stitched.nx"),
                entry="entry0000",
            )
            .get_identifier()
            .to_str()
        )
        self._mainWindow._stitchingConfigWindow.setPostProcessingOutput(
            HDF5Volume(
                file_path=os.path.join(
                    os.getcwd(), "stitching", "stitched_volume.hdf5"
                ),
                data_path="stitched_volume",
            )
            .get_identifier()
            .to_str()
        )

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Apply).clicked.connect(self.accept)
        self._loadConfigurationAction.triggered.connect(
            functools.partial(
                self._mainWindow._stitchingConfigWindow._loadSettings, file_path=None
            )
        )
        self._saveConfigurationAction.triggered.connect(
            functools.partial(
                self._mainWindow._stitchingConfigWindow._saveSettings, file_path=None
            )
        )
        self._addTomoObjectAction.triggered.connect(self._callbackAddTomoObj)

    def _callbackAddTomoObj(self, *args, **kwargs):
        """move interface to the z-ordered list and call the 'add tomo obj' callback"""
        orderedListWidget = (
            self._mainWindow._stitchingConfigWindow._widget._mainWidget._axisOrderedList
        )
        self._mainWindow._stitchingConfigWindow._widget._mainWidget.setCurrentWidget(
            orderedListWidget
        )
        tomo_objs = orderedListWidget._callbackAddTomoObj()
        for tomo_obj in tomo_objs:
            if len(tomo_objs) > 0:
                self.__updateStitchingTypeFromObjAdded(tomo_obj=tomo_objs[0])

    def __updateStitchingTypeFromObjAdded(self, tomo_obj: TomwerObject):
        if isinstance(tomo_obj, TomwerScanBase):
            self.setStitchingType(self.get_available_pre_processing_stitching_mode()[0])
        elif isinstance(tomo_obj, TomwerVolumeBase):
            self.setStitchingType(
                self.get_available_post_processing_stitching_mode()[0]
            )
        else:
            raise TypeError(
                f"tomo_obj is expected to be an instance of {type(TomwerObject)}. Got {type(tomo_obj)}"
            )

    def close(self):
        self._mainWindow.close()
        # requested for the waiting plot update
        super().close()

    def reject(self):
        self.close()
        super().reject()

    @classmethod
    def can_write_to_folder(cls, folder_path):
        if not os.path.exists(folder_path):
            parent_dir = os.path.dirname(folder_path)
            if (
                not parent_dir
            ):  # If folder_path is just a filename in the current directory
                parent_dir = os.getcwd()
            return cls.can_write_to_folder(parent_dir)
        return os.access(folder_path, os.W_OK)

    def __checkOutputDirWritable(self) -> bool:
        """Make sure the output directory is writable"""
        error = None
        config = self.getStitchingConfiguration()

        stitching_type = self.getStitchingType()
        if stitching_type in (StitchingType.Y_PREPROC, StitchingType.Z_PREPROC):
            output_file = config.get("preproc", {}).get("location", "")
            output_dir, _ = os.path.split(output_file)

        elif stitching_type in (StitchingType.Z_POSTPROC,):
            output_volume_identifier = config.get("postproc", {}).get(
                "output_volume", None
            )
            if output_volume_identifier:
                output_volume = VolumeFactory.create_tomo_object_from_identifier(
                    output_volume_identifier
                )
                if isinstance(output_volume, (HDF5Volume, MultiTIFFVolume)):
                    output_dir = os.path.dirname(output_volume.data_url.file_path())
                else:
                    output_dir = output_volume.data_url.file_path()
            else:
                output_dir = None
                error = "No output file/directory defined."
        else:
            raise ValueError(f"stitching type {stitching_type!r} not handled.")
        if output_dir is not None:
            if not self.can_write_to_folder(output_dir):
                error = f"No write rights on {output_dir!r}. Unable to save stitching."
        config.get("preproc")

        if error is None:
            return True

        error_dialog = qt.QMessageBox()
        error_dialog.setIcon(qt.QMessageBox.Warning)
        error_dialog.setWindowTitle("Error - cannot process")
        error_dialog.setText(error)
        error_dialog.setInformativeText("Please check your input and try again.")
        error_dialog.exec()
        return False

    def __checkStitchingModeSeemsValid(self) -> bool:
        """
        Check whether the requested stitching mode matches the input data (pre-processing or post-processing).
        If not, warn the user and block processing.
        """
        warning = None
        tomo_objs = self._mainWindow.getTomoObjs()
        stitching_type = self.getStitchingType()
        if len(tomo_objs) == 0:
            warning = "No object to be stitched together."
        else:
            scans = tuple(
                filter(
                    lambda obj: isinstance(obj, TomwerScanBase),
                    tomo_objs,
                )
            )
            volumes = tuple(
                filter(
                    lambda obj: isinstance(obj, TomwerVolumeBase),
                    tomo_objs,
                )
            )
            if len(scans) > 0 and len(volumes) > 0:
                warning = "Both scans and volumes to be stitched have been found. It should be one or the other. Processing will fail."

            elif stitching_type in (StitchingType.Y_PREPROC, StitchingType.Z_PREPROC):
                if len(scans) == 0:
                    warning = f"No scans found but {stitching_type.value!r} requested. Maybe you want to stitch reconstructed volumes instead of scans ? If so please select 'stitching method' to one of the 'postproc' mode."

            elif stitching_type in (StitchingType.Z_POSTPROC,):
                if len(volumes) == 0:
                    warning = f"No volumes found but {stitching_type.value!r} requested. Maybe you want to stitch scans instead of volumes ? If so please select 'stitching method' to one of the 'preproc' mode."
            else:
                raise NotImplementedError(
                    f"No consistency check is available for stitching type: {stitching_type.value}!)"
                )

        if warning is None:
            return True

        error_dialog = qt.QMessageBox()
        error_dialog.setIcon(qt.QMessageBox.Warning)
        error_dialog.setWindowTitle("Error - cannot process")
        error_dialog.setText(warning)
        error_dialog.setInformativeText("Please check your input and try again.")
        error_dialog.exec()
        return False

    def accept(self):
        if self.__checkStitchingModeSeemsValid() and self.__checkOutputDirWritable():
            self.process_stitching()

    def process_stitching(self):
        self._processingThread = _StitcherThread(
            parent=self,
            stitching_config=self.getStitchingConfiguration(),
            cluster_config=self.getClusterConfig(),
        )
        self._processingThread.start()

    def disableSlurmconfig(self):
        self._mainWindow._slurmCB.setChecked(False)
        self._mainWindow._slurmCB.setCheckable(False)

        slurm_config_idx = self._mainWindow.indexOf(self._mainWindow._slurmConfig)

        self._mainWindow.setTabEnabled(
            slurm_config_idx,
            False,
        )
        self._mainWindow.setTabToolTip(
            slurm_config_idx,
            "not available because unable to find slurm command (computer is probably not a slurm client ?",
        )

    # expose API
    def addTomoObj(self, *args, **kwargs):
        self._mainWindow.addTomoObj(*args, **kwargs)

    def getTomoObjs(self):
        return self._mainWindow._stitchingConfigWindow.getTomoObjs()

    def getClusterConfig(self) -> dict:
        if self._mainWindow._slurmCB.isChecked():
            return self._mainWindow.getClusterConfig()
        else:
            return {}

    def getStitchingConfiguration(self) -> dict:
        return self._mainWindow.getStitchingConfiguration()

    def setStitchingType(self, stitching_type):
        self._mainWindow.setStitchingType(stitching_type)

    def getStitchingType(self):
        return self._mainWindow.getStitchingType()

    def loadSettings(self, config_file: str):
        self._mainWindow.loadSettings(config_file)

    def setPreProcessingOutput(self, url: DataUrl) -> None:
        self._mainWindow._stitchingConfigWindow.setPreProcessingOutput(url=url)

    def setPostProcessingOutput(self, url: DataUrl) -> None:
        self._mainWindow._stitchingConfigWindow.setPostProcessingOutput(url=url)

    def get_available_pre_processing_stitching_mode(self) -> tuple[StitchingType]:
        return self._mainWindow.get_available_pre_processing_stitching_mode()

    def get_available_post_processing_stitching_mode(self) -> tuple[StitchingType]:
        return self._mainWindow.get_available_post_processing_stitching_mode()


def main(argv, stitcher_name: str, stitching_axis: int, logger):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "tomo_objs",
        help="All tomo objects to be stitched together",
        nargs="*",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    parser.add_argument(
        "--config",
        "--config-file",
        dest="config_file",
        default=None,
        help="Provide stitching configuration file to load parameters from it",
    )
    parser.add_argument(
        "--use-opengl-plot",
        "--opengl-backend",
        help="Use OpenGL for plots (instead of matplotlib)",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--check-scans",
        help="If True will check scans before adding them. This check scan validity (virtual dataset contains data...)",
        action="store_true",
        default=False,
    )
    options = parser.parse_args(argv[1:])

    if options.use_opengl_plot:
        silx.config.DEFAULT_PLOT_BACKEND = "gl"
    else:
        silx.config.DEFAULT_PLOT_BACKEND = "matplotlib"

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    window = MainWindow(axis=stitching_axis)
    window.setWindowFlags(qt.Qt.Window)
    window.setWindowTitle(stitcher_name)
    window.setWindowIcon(icons.getQIcon("tomwer"))

    window.show()
    from sluurp.utils import has_sbatch_available

    # load configuration is some provided
    if options.config_file is not None:
        window.loadSettings(options.config_file)

    if not has_sbatch_available():
        window.disableSlurmconfig()

    tomo_objs, (has_scans, has_vols) = get_tomo_objs_instances(options.tomo_objs)
    for tomo_obj in tomo_objs:
        if isinstance(tomo_obj, TomwerScanBase):
            tomo_obj.set_check_behavior(
                run_check=options.check_scans
            )  # speed up processing
        window.addTomoObj(tomo_obj=tomo_obj)
    if has_scans and has_vols:
        logger.warning(
            "seems like you have both volumes and scan on your inputs. Unable to stitch the two at the same time"
        )
    elif has_scans:
        if stitching_axis == 0:
            window.setStitchingType(StitchingType.Z_PREPROC)
        elif stitching_axis == 1:
            window.setStitchingType(StitchingType.Y_PREPROC)
        else:
            raise ValueError(
                f"stitching on scans is not possible along the axis {stitching_axis} (must be 0 or 1)"
            )
    elif has_vols:
        if stitching_axis == 0:
            window.setStitchingType(StitchingType.Z_POSTPROC)
        else:
            raise ValueError(
                f"stitching on volumes is not possible along the axis {stitching_axis} (must be 0). Should be possible in theory. Please contact us so we can have a dataset and test it"
            )

    window.setAttribute(qt.Qt.WA_DeleteOnClose)
    splash.finish(window)
    exit(app.exec())


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.closeAllWindows()  # needed because get a waiting thread behind
    qt.QApplication.quit()



from __future__ import annotations


import os
import pint
import tempfile

from tomoscan.series import Series

from tomwer.gui.stitching.StitchingWindow import ZStitchingWindow
from tomwer.gui.stitching.tests.utils import create_scans_z_series
from tomwer.tests.conftest import qtapp  # noqa F401


_ureg = pint.get_application_registry()


def test_ZStitchingWindow(
    qtapp,  # noqa F811
    tmp_path,
):
    """
    Test ZStitchingWindow

    * adding tom objects
    * saving and loading settings (nabu-stitching configuration)
    * editing tomo objects positions over stitching axis
    * reset tomo objects positions
    """

    window = ZStitchingWindow()

    axis_0_positions = (90, 0.0, -90.0)
    axis_2_positions = (0.0, 0.0, 0.0)
    pixel_size = 1.0

    input_dir = os.path.join(tmp_path, "input")

    scans = create_scans_z_series(
        output_dir=input_dir,
        z_positions_m=axis_0_positions,
        x_positions_m=axis_2_positions,
        shifts=((0.0, 0.0), (-90.0, 0.0), (-180.0, 0.0)),
        sample_pixel_size=pixel_size,
        raw_frame_width=280,
        final_frame_width=100,
    )
    series = Series("z-series", scans)

    window.show()
    for scan in scans:
        window.addTomoObj(scan)
    window.clean()
    window.setSeries(series)

    # test dumping and loading configuration to a file
    with tempfile.TemporaryDirectory() as dump_dir:
        config_file = os.path.join(dump_dir, "configuration.cfg")
        window._saveSettings(file_path=config_file)
        assert os.path.exists(config_file)
        window._loadSettings(config_file)
        # remove configuration
        window.clean()
        assert len(window._widget._mainWidget.getTomoObjs()) == 0
        # reload it
        window._loadSettings(file_path=config_file)
        assert len(window._widget._mainWidget.getTomoObjs()) == len(series)

    assert window.getConfiguration()["stitching"]["axis_0_pos_px"] == [90, 0, -90]

    # test editing axis position
    widget_tomo_obj_0 = (
        window._editTomoObjFirstAxisPositionsWidget._tomoObjtoTomoObjPosWidget[
            scans[0].get_identifier().to_str()
        ]
    )

    window._editTomoObjFirstAxisPositionsWidget.setEditionMode("free")
    widget_tomo_obj_0.setValue(10, emit_editing_finished=True)
    assert window.getConfiguration()["stitching"]["axis_0_pos_px"] == [10, 0, -90]
    window._editTomoObjFirstAxisPositionsWidget.setEditionMode("downstream")
    widget_tomo_obj_0.setValue(30, emit_editing_finished=True)
    assert window.getConfiguration()["stitching"]["axis_0_pos_px"] == [30, 20, -70]
    window._editTomoObjFirstAxisPositionsWidget.setEditionMode("upstream")
    widget_tomo_obj_2 = (
        window._editTomoObjFirstAxisPositionsWidget._tomoObjtoTomoObjPosWidget[
            scans[2].get_identifier().to_str()
        ]
    )
    widget_tomo_obj_2.setValue(-120, emit_editing_finished=True)
    assert window.getConfiguration()["stitching"]["axis_0_pos_px"] == [-20, -30, -120]

    # test reset positions to initial positions
    window._editTomoObjFirstAxisPositionsWidget._resetPositions()
    assert window.getConfiguration()["stitching"]["axis_0_pos_px"] == [90, 0, -90]



from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.reconstruction.nabu.castvolume import CastVolumeWidget


def test_CastVolumeWidget(qtapp):  # noqa F811
    """simple test of the CastVolumeWidget"""
    widget = CastVolumeWidget(parent=None)
    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": None,
        "data_min": None,
        "output_data_type": "uint16",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "tiff",
        "overwrite": True,
        "rescale_max_percentile": 90,
        "rescale_min_percentile": 10,
        "remove_input_volume": False,
    }

    widget._minMaxAuto.setChecked(False)
    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": 0.0,
        "data_min": 0.0,
        "output_data_type": "uint16",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "tiff",
        "overwrite": True,
        "rescale_max_percentile": None,
        "rescale_min_percentile": None,
        "remove_input_volume": False,
    }

    widget.setDataMin(1.0)
    widget.setDataMax(10.0)

    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": 10.0,
        "data_min": 1.0,
        "output_data_type": "uint16",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "tiff",
        "overwrite": True,
        "rescale_max_percentile": None,
        "rescale_min_percentile": None,
        "remove_input_volume": False,
    }
    widget.setOutputFileformat("edf")
    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": 10.0,
        "data_min": 1.0,
        "output_data_type": "uint16",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "edf",
        "overwrite": True,
        "rescale_max_percentile": None,
        "rescale_min_percentile": None,
        "remove_input_volume": False,
    }
    widget.setOverwrite(False)
    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": 10.0,
        "data_min": 1.0,
        "output_data_type": "uint16",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "edf",
        "overwrite": False,
        "rescale_max_percentile": None,
        "rescale_min_percentile": None,
        "remove_input_volume": False,
    }
    widget.setOutputDataType("float32")
    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": 10.0,
        "data_min": 1.0,
        "output_data_type": "float32",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "edf",
        "overwrite": False,
        "rescale_max_percentile": None,
        "rescale_min_percentile": None,
        "remove_input_volume": False,
    }
    widget.setRemoveInputVolume(True)
    assert widget.getConfiguration() == {
        "compression_ratios": None,
        "data_max": 10.0,
        "data_min": 1.0,
        "output_data_type": "float32",
        "output_dir": "{volume_data_parent_folder}/cast_volume",
        "output_file_format": "edf",
        "overwrite": False,
        "rescale_max_percentile": None,
        "rescale_min_percentile": None,
        "remove_input_volume": True,
    }



from __future__ import annotations

from silx.gui import qt
from tomwer.core.volume.volumebase import TomwerVolumeBase
from weakref import ref


class SliceLoader(qt.QThread):
    """
    Thread to load a single slice from a volume.
    """

    def __init__(
        self, parent, volume: TomwerVolumeBase, slice_index: int, axis: int
    ) -> None:
        super().__init__(parent)
        if not isinstance(volume, TomwerVolumeBase):
            raise TypeError()
        self.__volume = ref(volume)
        self.slice_index = slice_index
        self.__axis = axis
        self.__data = None

    def run(self):
        if self.__volume() is None:
            return
        self.__data = self.__volume().get_slice(
            index=self.slice_index,
            axis=self.__axis,
        )

    @property
    def data(self):
        return self.__data



"""
Button of general usage.
"""

from __future__ import annotations

import pint
import numpy
import logging
from silx.gui import qt
from silx.gui.plot.tools.roi import RegionOfInterestManager
from silx.gui.plot.items.roi import LineROI
from silx.gui.plot.PlotToolButtons import PlotToolButton

from tomwer.gui import icons

_logger = logging.getLogger(__file__)

_ureg = pint.get_application_registry()


class PadlockButton(qt.QPushButton):
    """Simple button to define a button with PadLock icons"""

    sigLockChanged = qt.Signal(bool)
    """signal emitted when the lock status change"""

    def __init__(self, parent):
        qt.QPushButton.__init__(self, parent)
        self._lockIcon = icons.getQIcon("locked")
        self._unlockIcon = icons.getQIcon("unlocked")
        self.setIcon(self._unlockIcon)
        self.setCheckable(True)

        # connect signals
        self.toggled.connect(self._updateDisplay)

    def setLock(self, lock: bool):
        self.setChecked(lock)
        self._updateDisplay(lock)

    def _updateDisplay(self, checked):
        _icon = self._lockIcon if checked else self._unlockIcon
        self.setIcon(_icon)
        self.sigLockChanged.emit(checked)

    def isLocked(self) -> bool:
        return self.isChecked()


class TabBrowsersButtons(qt.QWidget):
    """Simple widget containing buttons to go to 'next' or 'previous'"""

    sigPreviousReleased = qt.Signal()
    """emitted when the previous button is released"""
    sigNextReleased = qt.Signal()
    """emitted when the next button is released"""

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QHBoxLayout())
        # define gui
        style = qt.QApplication.style()
        self._previousButton = qt.QPushButton("previous", self)
        previous_icon = style.standardIcon(qt.QStyle.SP_ArrowLeft)
        self._previousButton.setIcon(previous_icon)
        self.layout().addWidget(self._previousButton)

        self._nextButton = qt.QPushButton("next", self)
        next_icon = style.standardIcon(qt.QStyle.SP_ArrowRight)
        self._nextButton.setIcon(next_icon)
        self.layout().addWidget(self._nextButton)
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self.layout().addWidget(spacer)

        # connect signal / slot
        self._previousButton.released.connect(self._previousReleased)
        self._nextButton.released.connect(self._nextReleased)

    def _previousReleased(self, *args, **kwargs):
        self.sigPreviousReleased.emit()

    def _nextReleased(self, *args, **kwargs):
        self.sigNextReleased.emit()


class TapeMeasureToolButton(PlotToolButton):
    """Button to active measurement between two point of the plot"""

    def __init__(self, parent=None, plot=None, pixel_size: pint.Quantity | None = None):
        super().__init__(parent=parent, plot=plot)
        self._roiManager = None
        self._lastRoiCreated = None
        self._pixel_size: pint.Quantity | None = pixel_size
        self.setIcon(icons.getQIcon("ruler"))
        self.setToolTip("measure distance between two pixels")
        self.toggled.connect(self._callback)
        self._connectPlot(plot)

    def setPlot(self, plot):
        return super().setPlot(plot)

    def setPixelSize(self, pixel_size: tuple[pint.Quantity, pint.Quantity] | None):
        self._pixel_size = pixel_size

    def _callback(self, toggled):
        if not self._roiManager:
            return
        if self._lastRoiCreated is not None:
            self._lastRoiCreated.setVisible(self.isChecked())
        if self.isChecked():
            self._roiManager.start(
                TapeMeasureROI,
                self,
            )
            self.__interactiveModeStarted(self._roiManager)
        else:
            source = self._roiManager.getInteractionSource()
            if source is self:
                self._roiManager.stop()

    def __interactiveModeStarted(self, roiManager):
        roiManager.sigInteractiveModeFinished.connect(self.__interactiveModeFinished)

    def __interactiveModeFinished(self):
        roiManager = self._roiManager
        if roiManager is not None:
            roiManager.sigInteractiveModeFinished.disconnect(
                self.__interactiveModeFinished
            )
        self.setChecked(False)

    def _connectPlot(self, plot):
        """
        Called when the plot is connected to the widget
        :param plot: :class:`.PlotWidget` instance
        """
        if plot is None:
            return
        self._roiManager = RegionOfInterestManager(plot)
        self._roiManager.setColor("yellow")  # Set the color of ROI
        self._roiManager.sigRoiAdded.connect(self._registerCurrentROI)

    def _disconnectPlot(self, plot):
        if plot and self._lastRoiCreated is not None:
            self._roiManager.removeRoi(self._lastRoiCreated)
            self._lastRoiCreated = None
        return super()._disconnectPlot(plot)

    def _registerCurrentROI(self, currentRoi):
        if self._lastRoiCreated is None:
            self._lastRoiCreated = currentRoi
            self._lastRoiCreated.setPixelSize(self._pixel_size)
        elif currentRoi != self._lastRoiCreated and self._roiManager is not None:
            self._roiManager.removeRoi(self._lastRoiCreated)
            self._lastRoiCreated = currentRoi
            self._lastRoiCreated.setPixelSize(self._pixel_size)


class TapeMeasureROI(LineROI):
    """ROI dedicated to tape measure"""

    def __init__(self, parent=None, pixel_size: pint.Quantity | None = None):
        super().__init__(parent)
        self._pixel_size: pint.Quantity | None = None
        self.setPixelSize(pixel_size)

    def setEndPoints(self, startPoint, endPoint):
        distance_px = numpy.linalg.norm(endPoint - startPoint)
        super().setEndPoints(startPoint=startPoint, endPoint=endPoint)
        if self._pixel_size is None:
            self._updateText(f"{distance_px :.1f}px")
        else:
            distance_m = distance_px * self._pixel_size
            value = self.cast_metric_to_best_unit(distance_m)
            self._updateText(f"{value:.4f~P}")

    def setPixelSize(self, pixel_size: tuple[pint.Quantity]):
        if isinstance(pixel_size, (tuple, list)):
            assert (
                len(pixel_size) == 2
            ), "expects at most two pixel size values (x and y values)"
            if not numpy.isclose(
                pixel_size[0].to_base_units(), pixel_size[1].to_base_units()
            ):
                value = self.cast_metric_to_best_unit(pixel_size[0])
                _logger.warning(
                    f"TapeMeasure is only handling square pixels for now. Will consider the pixel is {value:.4f~P}x{value:.4f~P}"
                )
            pixel_size = pixel_size[0]
        self._pixel_size = pixel_size

    @staticmethod
    def cast_metric_to_best_unit(value: pint.Quantity) -> pint.Quantity:
        value = value.to_base_units()
        if value < (1.0 * _ureg.micrometer):
            return value.to(_ureg.nanometer)
        elif value < (0.1 * _ureg.millimeter):  # prefer mm to um
            return value.to(_ureg.micrometer)
        elif value < (1.0 * _ureg.centimeter):
            return value.to(_ureg.millimeter)
        elif value < (1.0 * _ureg.meter):
            return value.to(_ureg.centimeter)
        else:
            return value.to(_ureg.meter)



from __future__ import annotations

from silx.gui import qt


class QThreadListenerBase(qt.QThread):
    """
    This class defines the API for a QThread listening to blissdata.

    Interface is defined by signals emitted and the constructor signature.
    The goal is to be able to connect it with the DataListenerWidget
    """

    sigAcquisitionStarted = qt.Signal(tuple)
    """Signal emitted when an acquisition is started. Tuple is:
    (master_file, master_entry)"""
    sigAcquisitionEnded = qt.Signal(tuple)
    """Signal emitted when an acquisition is ended. Tuple is
    (master_file, master_entry, succeed)"""
    sigScanAdded = qt.Signal(tuple)
    """Signal emitted when a scan is added to an acquisition. Tuple is
    (master_file, master_entry, scan_entry)"""
    sigServerStop = qt.Signal()
    """Signal if the rpc-server have been turn off"""

    def __init__(self, host, port, acquisitions):
        super().__init__()

    def join(self, timeout=None):
        if timeout is None:
            self.wait()
        else:
            self.wait(timeout)

    def stop(self):
        raise NotImplementedError("Base class")



from __future__ import annotations


import numpy

from tomoscan.volumebase import SliceTuple

from tomwer.tests.conftest import qtapp  # noqa F401
from tomwer.gui.visualization.volume_viewer.VolumeViewerWidget import (
    VolumeViewerWidget,
)


def test_volumeReconstructionSummary(qtapp):  # noqa F811

    widget = VolumeViewerWidget()
    widget.setSlices(
        {
            SliceTuple(axis=0, index=1): numpy.ones((10, 20)),
            SliceTuple(axis=0, index=12): numpy.ones((10, 20)),
            SliceTuple(axis=0, index=25): numpy.ones((10, 20)),
            SliceTuple(axis=1, index=0): numpy.arange(0, 100).reshape(10, 10),
            SliceTuple(axis=1, index=10): numpy.arange(1000, 1100).reshape(10, 10),
            SliceTuple(axis=1, index=20): numpy.arange(1000, 1100).reshape(10, 10),
            SliceTuple(axis=2, index=100): numpy.random.random(100).reshape(10, 10),
            SliceTuple(axis=2, index=101): numpy.random.random(100).reshape(10, 10),
            SliceTuple(axis=2, index=103): numpy.random.random(100).reshape(10, 10),
        }
    )
    widget.setVolumeMetadata(
        {
            "nabu_config": {
                "reconstruction": {
                    "method": "FBP",
                },
                "phase": {
                    "method": "paganin",
                    "delta_beta": 110.0,
                },
            },
            "processing_options": {
                "reconstruction": {
                    "voxel_size_cm": (0.2, 0.2, 0.2),
                    "rotation_axis_position": 104,
                    "enable_halftomo": True,
                    "fbp_filter_type": "Hilbert",
                    "sample_detector_dist": 0.4,
                },
                "take_log": {
                    "log_min_clip": 1.0,
                    "log_max_clip": 10.0,
                },
            },
        },
        volume_shape=(10, 10, 10),
    )



"""utils for orange widget"""

from __future__ import annotations


class WidgetLongProcessing:
    """Class to display processing for some widgets with long processing"""

    def processing_state(self, working: bool, info=None) -> None:
        """

        :param working:
        :param info:
        """
        # default orange version don't have Processing.
        try:
            if working:
                self.Processing._add_general(
                    info or "processing", text=None, shown=True
                )
            else:
                try:
                    self.Processing.clear()
                except Exception:
                    pass
        except AttributeError:
            # in case we are on an orange version not having `Processing`
            pass

    def setDryRun(self, dry_run):
        pass

    def _startProcessing(self, *args, **kwargs):
        self.processing_state(working=True, info="processing")

    def _endProcessing(self, *args, **kwargs):
        self.processing_state(working=False, info="processing")



from __future__ import annotations

import pint
from silx.gui import qt
from processview.gui.DropDownWidget import DropDownWidget
from tomwer.synctools.axis import QAxisRP
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.reconstruction.axis.EstimatedCorComboBox import EstimatedCorComboBox
from tomwer.core.process.reconstruction.axis.side import Side
from tomwer.core.process.reconstruction.axis import mode as axis_mode
from tomwer.gui.fonts import FONT_SMALL
from tomwer.gui.utils.scrollarea import QDoubleSpinBoxIgnoreWheel

_ureg = pint.get_application_registry()


class EstimatedCORWidget(qt.QGroupBox):
    """
    Widget to define the estimated center of rotation.
    (based on the motor offset and the 'x_rotation_axis_pixel_position')
    """

    sigValueChanged = qt.Signal()
    """Emit when one of the value changed"""
    sigUpdateXRotAxisPixelPosOnNewScan = qt.Signal()
    """Emit when user want to stop / activate x rotation axis pixel position when a new scan arrives"""
    sigYAxisInvertedChanged = qt.Signal(bool)

    def __init__(self, parent, axis_params: QAxisRP):
        self._axis_params = axis_params
        self._imageWidth = None

        super().__init__(parent)
        self.setLayout(qt.QGridLayout())
        # estimated cor
        self._estimatedCORLabel = qt.QLabel("Estimated CoR (relative)", self)
        self.layout().addWidget(self._estimatedCORLabel, 0, 0, 1, 1)

        self._estimatedCORQCB = EstimatedCorComboBox(self)
        self.layout().addWidget(self._estimatedCORQCB, 0, 1, 1, 1)

        # offset calibration
        self._offsetWidgetDropdown = DropDownWidget(
            parent=self, direction=qt.Qt.LayoutDirection.RightToLeft
        )
        self._offsetWidget = _OffsetCalibration(parent=self, axis_params=axis_params)
        self.layout().addWidget(self._offsetWidgetDropdown, 1, 0, 2, 2)
        self._offsetWidgetDropdown.setWidget(self._offsetWidget)

        # set up
        self._offsetWidgetDropdown.setChecked(False)

        # connect signal / slot
        self._estimatedCORQCB.sigEstimatedCorChanged.connect(self._corChanged)
        self._offsetWidget.sigOffsetChanged.connect(
            self._updateEstimatedCorFromMotorOffsetWidget
        )
        self._offsetWidget.sigXRotationAxisPixelPositionChanged.connect(
            self._updateEstimatedCorFromMotorOffsetWidget
        )
        self._offsetWidget.sigUpdateXRotAxisPixelPosOnNewScan.connect(
            self.sigUpdateXRotAxisPixelPosOnNewScan
        )
        self._offsetWidget.sigYAxisInvertedChanged.connect(self.sigYAxisInvertedChanged)

    def getEstimatedCor(self):
        return self._estimatedCORQCB.getCurrentCorValue()

    def setEstimatedCor(
        self, value: float | Side | str, provided_with_offset: bool = False
    ):

        if isinstance(value, float):
            if self._offsetWidget.isYAxisInverted():
                value = -1.0 * value

            if provided_with_offset:
                value_with_offset = value
                value_without_offset = value - self._offsetWidget.getOffset()
            else:
                value_with_offset = value + self._offsetWidget.getOffset()
                value_without_offset = value
            with block_signals(self._offsetWidget):
                self._offsetWidget.setXRotationAxisPixelPosition(
                    value_without_offset, apply_flip=False
                )
        else:
            # case this is a side
            value_with_offset = value

        self._estimatedCORQCB.setCurrentCorValue(value_with_offset)

        with block_signals(self._axis_params):
            self._axis_params.estimated_cor = value_with_offset

    def _corChanged(self, value):
        assert value is None or isinstance(value, (float, Side, None))
        with block_signals(self._axis_params):
            self._axis_params.estimated_cor = value
        self.sigValueChanged.emit()

    def _updateVisibleSides(self, mode: axis_mode.AxisMode):
        """
        Update the visibility and selection of sides (Left, Center, Right)
        for the Estimated Center of Rotation (CoR) ComboBox based on the
        provided axis mode and a calculated CoR guess.

        This method adjusts the available sides and determines the new
        side (Left, Center, or Right) based on the position of a CoR guess
        relative to the width of the image. The image is divided into thirds:
        - Left: From -infinity to the first third.
        - Center: From the first third to the second third.
        - Right: From the second third to infinity.

        If the CoR guess is not valid or falls outside the calculated bounds,
        a default valid side is selected.

        Behavior:
        ---------
        - By default if the method allows it, the side will be the estimated CoR.
        - If the method does not allow it, it will update the visible sides in
        the Estimated CoR ComboBox based on the valid sides defined in the axis
        mode metadata and dynamically determines the new side (Left, Center, Right)
        based on the CoR guess position within the image width.

        """
        mode = axis_mode.AxisMode.from_value(mode)
        valid_sides = axis_mode.AXIS_MODE_METADATAS[mode].valid_sides
        self._estimatedCORQCB.setSidesVisible(valid_sides)
        first_guess_available = axis_mode.AXIS_MODE_METADATAS[
            mode
        ].allows_estimated_cor_as_numerical_value
        self._estimatedCORQCB.setFirstGuessAvailable(first_guess_available)

        if first_guess_available:
            # if the first guess is valid, when the sides visibility is modify we want to activate it.
            self._estimatedCORQCB.selectFirstGuess()
        elif valid_sides:
            # Proceed only if there are valid sides
            current_side = self._estimatedCORQCB.getCurrentCorValue()
            if not isinstance(current_side, Side) or current_side not in valid_sides:
                with block_signals(self._estimatedCORQCB):
                    cor_guess = self._estimatedCORQCB.getCurrentCorValue()
                    if cor_guess is not None and self._imageWidth is not None:
                        left_boundary = -float("inf")
                        right_boundary = float("inf")
                        middle_left_boundary = (
                            self._imageWidth / 3 - self._imageWidth / 2
                        )
                        middle_right_boundary = (
                            2 * self._imageWidth / 3 - self._imageWidth / 2
                        )

                        if (
                            Side.LEFT in valid_sides
                            and left_boundary <= cor_guess < middle_left_boundary
                        ):
                            new_side = Side.LEFT
                        elif (
                            Side.CENTER in valid_sides
                            and middle_left_boundary
                            <= cor_guess
                            < middle_right_boundary
                        ):
                            new_side = Side.CENTER
                        elif (
                            Side.RIGHT in valid_sides
                            and middle_right_boundary <= cor_guess <= right_boundary
                        ):
                            new_side = Side.RIGHT
                        else:
                            # Fallback to the first available valid side
                            new_side = valid_sides[0]
                    else:
                        # If no guess or boundaries are available, fallback to the first valid side
                        new_side = valid_sides[0]
                    self._estimatedCORQCB.setCurrentCorValue(new_side)
                    self._axis_params.estimated_cor = new_side

    def _updateEstimatedCorFromMotorOffsetWidget(self):
        self._estimatedCORQCB.setCurrentCorValue(self._offsetWidget.getEstimatedCor())
        self.sigValueChanged.emit()

    def setImageWidth(self, image_width: float | None):
        self._imageWidth = image_width

    # expose API
    def updateXRotationAxisPixelPositionOnNewScan(self) -> bool:
        return self._offsetWidget.updateXRotationAxisPixelPositionOnNewScan()

    def setUpdateXRotationAxisPixelPositionOnNewScan(self, update: bool):
        self._offsetWidget.setUpdateXRotationAxisPixelPositionOnNewScan(update=update)

    def setPixelSize(self, pixel_size_m: float | None) -> None:
        self._offsetWidget.setPixelSize(pixel_size_m=pixel_size_m)

    def isYAxisInverted(self) -> bool:
        return self._offsetWidget.isYAxisInverted()

    def setYAxisInverted(self, checked: bool):
        return self._offsetWidget.setYAxisInverted(checked=checked)


class _OffsetCalibration(qt.QGroupBox):

    sigOffsetChanged = qt.Signal()
    sigXRotationAxisPixelPositionChanged = qt.Signal()
    sigUpdateXRotAxisPixelPosOnNewScan = qt.Signal()
    sigYAxisInvertedChanged = qt.Signal(bool)

    def __init__(self, parent, axis_params: QAxisRP):
        super().__init__(parent)
        self._axis_params = axis_params

        self.setLayout(qt.QVBoxLayout())

        # x_rotation_axis_pixel_position
        self._xRotationAxisPixelPositionGroup = qt.QGroupBox("NXtomo metadata ")
        self._xRotationAxisPixelPositionGroup.setLayout(qt.QGridLayout())
        self.layout().addWidget(self._xRotationAxisPixelPositionGroup)
        self._xRotationAxisPixelPositionLabel = qt.QLabel(
            "x_rotation_axis_pixel_position"
        )
        self._xRotationAxisPixelPositionGroup.layout().addWidget(
            self._xRotationAxisPixelPositionLabel, 0, 0, 2, 1
        )
        self._xRotationAxisPixelPositionDSB = QDoubleSpinBoxIgnoreWheel(self)
        self._xRotationAxisPixelPositionDSB.setDecimals(2)
        self._xRotationAxisPixelPositionDSB.setRange(-float("inf"), float("inf"))
        self._xRotationAxisPixelPositionDSB.setSuffix(" px")
        self._xRotationAxisPixelPositionDSB.setEnabled(False)
        self._xRotationAxisPixelPositionGroup.layout().addWidget(
            self._xRotationAxisPixelPositionDSB, 0, 1, 2, 1
        )
        self._xRotationAxisPixelPositionKeepUpdatedCB = qt.QCheckBox(
            "Update with\n new scan"
        )
        self._xRotationAxisPixelPositionKeepUpdatedCB.setFont(FONT_SMALL)
        self._xRotationAxisPixelPositionKeepUpdatedCB.setToolTip(
            "Updates the value when a new scan arrives.\n"
            "Once updated this will change the numerical value of the estimated cor (from estimated relative cor)"
        )
        self._xRotationAxisPixelPositionGroup.layout().addWidget(
            self._xRotationAxisPixelPositionKeepUpdatedCB, 0, 2, 1, 1
        )
        self._yAxisInvertedCB = qt.QCheckBox("Y axis inverted")
        self._yAxisInvertedCB.setFont(FONT_SMALL)
        self._yAxisInvertedCB.setToolTip(
            "Sometime the y axis can be inverted (like on ID11).\nIn this case the estimation of the CoR has a flip that must be handled downstream"
        )
        self._xRotationAxisPixelPositionGroup.layout().addWidget(
            self._yAxisInvertedCB, 1, 2, 1, 1
        )

        # offset group
        self._offsetGroup = qt.QGroupBox("Custom Offset")
        self._offsetGroup.setLayout(qt.QFormLayout())
        self.layout().addWidget(self._offsetGroup)

        # offset
        self._offsetSB = QDoubleSpinBoxIgnoreWheel(self)
        self._offsetSB.setSuffix(" px")
        self._offsetSB.setDecimals(2)
        self._offsetSB.setRange(-float("inf"), float("inf"))
        self._offsetSB.setEnabled(False)
        self._offsetGroup.layout().addRow("Offset", self._offsetSB)

        # motor offset
        self._motorOffsetSB = QDoubleSpinBoxIgnoreWheel(self)
        self._motorOffsetSB.setSuffix(" mm")
        self._motorOffsetSB.setDecimals(4)
        self._motorOffsetSB.setRange(-float("inf"), float("inf"))
        self._offsetGroup.layout().addRow("Y Motor Offset", self._motorOffsetSB)

        # pixel size
        self._pixelSizeSB = QDoubleSpinBoxIgnoreWheel(self)
        self._pixelSizeSB.setSuffix(" µm")
        self._pixelSizeSB.setDecimals(3)
        self._pixelSizeSB.setRange(0, float("inf"))

        # Add a horizontal layout for the pixel size and the padlock
        pixelSizeLayout = qt.QHBoxLayout()

        # Add the spin box to the layout
        pixelSizeLayout.addWidget(self._pixelSizeSB)

        # Add a padlock button
        self._pixelSizePadlock = PadlockButton(self)
        self._pixelSizePadlock.setChecked(True)  # Default to locked
        self._pixelSizePadlock.setFixedSize(24, 24)
        self._pixelSizePadlock.setToolTip("Lock/Unlock pixel size")
        pixelSizeLayout.addWidget(self._pixelSizePadlock)

        # Add the layout to the offset group
        self._offsetGroup.layout().addRow("Pixel Size", pixelSizeLayout)

        # Initialize the spin box as disabled (locked by default)
        self._pixelSizeSB.setEnabled(False)

        # Connect signal to handle locking behavior
        self._pixelSizePadlock.toggled.connect(self._togglePixelSizeLock)

        # set up
        self._xRotationAxisPixelPositionKeepUpdatedCB.setChecked(True)
        self._yAxisInvertedCB.setChecked(False)

        # connect signal / slot
        self._offsetSB.editingFinished.connect(self._offsetEdited)
        self._xRotationAxisPixelPositionKeepUpdatedCB.toggled.connect(
            self.sigUpdateXRotAxisPixelPosOnNewScan
        )
        self._motorOffsetSB.valueChanged.connect(self._motorOffsetChanged)
        self._pixelSizeSB.valueChanged.connect(self._pixelSizeChanged)
        self._yAxisInvertedCB.toggled.connect(self._yAxisInverted)

    def getXRotationAxisPixelPosition(self) -> float:
        return self._xRotationAxisPixelPositionDSB.value()

    def updateXRotationAxisPixelPositionOnNewScan(self) -> bool:
        return self._xRotationAxisPixelPositionKeepUpdatedCB.isChecked()

    def setUpdateXRotationAxisPixelPositionOnNewScan(self, update):
        self._xRotationAxisPixelPositionKeepUpdatedCB.setChecked(update)

    def setXRotationAxisPixelPosition(
        self, value: float, apply_flip: bool = True
    ) -> float:
        """Set the 'x_rotation_axis_pixel_position' and flip it if necessary"""
        if apply_flip and self.isYAxisInverted():
            value = -1.0 * value
        self._xRotationAxisPixelPositionDSB.setValue(value)
        return value

    def getOffset(self) -> float:
        return self._offsetSB.value()

    def setOffset(self, value: float) -> None:
        self._offsetSB.setValue(value)
        self._offsetEdited()

    def _offsetEdited(self):
        with block_signals(self):
            self._axis_params.x_rotation_axis_pos_px_offset = self.getOffset()
        self.sigOffsetChanged.emit()

    def _motorOffsetChanged(self):
        """Recalculate the offset when the motor offset changes."""
        if self._pixelSizeSB.value() and self._motorOffsetSB.value():
            pixel_size = self._pixelSizeSB.value() * _ureg.micrometer
            motor_offset = self._motorOffsetSB.value() * _ureg.millimeter
            with block_signals(self):
                self.setOffset((motor_offset / pixel_size).to_base_units().magnitude)
            self.sigOffsetChanged.emit()

    def _pixelSizeChanged(self):
        """Recalculate the offset when the pixel size changes."""
        if self._pixelSizeSB.value() and self._motorOffsetSB.value():
            pixel_size = self._pixelSizeSB.value() * _ureg.micrometer
            motor_offset = self._motorOffsetSB.value() * _ureg.millimeter
            with block_signals(self):
                self.setOffset((motor_offset / pixel_size).to_base_units().magnitude)
            self.sigOffsetChanged.emit()

    def _xRotationAxisPixelPositionEdited(self):
        with block_signals(self._axis_params):
            self._axis_params.x_rotation_axis_pixel_position = (
                self.getXRotationAxisPixelPosition()
            )
        self.sigXRotationAxisPixelPositionChanged.emit()

    def _yAxisInverted(self):
        self.setXRotationAxisPixelPosition(
            value=-1.0 * self.getXRotationAxisPixelPosition(),
            apply_flip=False,
        )
        self._xRotationAxisPixelPositionEdited()

    def isYAxisInverted(self) -> bool:
        return self._yAxisInvertedCB.isChecked()

    def setYAxisInverted(self, checked: bool):
        self._yAxisInvertedCB.setChecked(checked)

    def getEstimatedCor(self) -> float:
        return self.getXRotationAxisPixelPosition() + self.getOffset()

    def setPixelSize(self, pixel_size_m: float | None):
        if pixel_size_m is None:
            self._pixelSizeSB.clear()
        else:
            self._pixelSizeSB.setValue(
                (pixel_size_m * _ureg.meter).to(_ureg.micrometer).magnitude
            )

    def _togglePixelSizeLock(self, locked: bool):
        """Lock or unlock the pixel size spin box."""
        self._pixelSizeSB.setEnabled(not locked)



# coding: utf-8
from __future__ import annotations


import gc

import pickle

from silx.gui import qt

from tomwer.tests.conftest import qtapp  # noqa F401
from orangecontrib.tomwer.widgets.deprecated.DataListenerOW import DataListenerOW
from orangecanvas.scheme.readwrite import literal_dumps


def test_DataListenerOW(
    qtapp,  # noqa F811
):

    widget = DataListenerOW(host_discovery=False)

    pickle.dumps(widget.get_configuration())
    literal_dumps(widget.get_configuration())

    widget.setAttribute(qt.Qt.WA_DeleteOnClose)
    widget.close()

    gc.collect()



from __future__ import annotations

import os
import pytest
import numpy
from tomwer.tests.conftest import nxtomo_scan_180  # noqa F811
from tomwer.core.process.drac.rawdataset import DracRawDataset


def test_IcatRawDataset(nxtomo_scan_180):  # noqa F811
    """
    test CreateRawDataScreenshotsTask task
    """
    # test nothing requested
    dataset = DracRawDataset(
        tomo_obj=nxtomo_scan_180,
        raw_projections_required=False,
        raw_darks_required=False,
        raw_flats_required=False,
    )

    assert not os.path.exists(dataset.get_gallery_dir())
    dataset.build_gallery()
    assert not os.path.exists(dataset.get_gallery_dir())

    # test only flat requested
    dataset = DracRawDataset(
        tomo_obj=nxtomo_scan_180,
        raw_projections_required=False,
        raw_darks_required=False,
        raw_flats_required=True,
    )
    dataset.build_gallery()
    assert len(os.listdir(dataset.get_gallery_dir())) == 1

    # test only dark requested
    dataset = DracRawDataset(
        tomo_obj=nxtomo_scan_180,
        raw_projections_required=False,
        raw_darks_required=True,
        raw_flats_required=False,
    )
    dataset.gallery_overwrite = False
    with pytest.raises(RuntimeError):
        dataset.build_gallery()
    dataset.gallery_overwrite = True
    dataset.build_gallery()
    assert len(os.listdir(dataset.get_gallery_dir())) == 1

    # test only projection requested
    dataset = DracRawDataset(
        tomo_obj=nxtomo_scan_180,
        raw_projections_required=True,
        raw_darks_required=False,
        raw_flats_required=False,
    )
    dataset.gallery_overwrite = True
    dataset.build_gallery()
    assert len(os.listdir(dataset.get_gallery_dir())) == 2

    # test all requested
    dataset = DracRawDataset(
        tomo_obj=nxtomo_scan_180,
        raw_projections_required=True,
        raw_darks_required=True,
        raw_flats_required=True,
        raw_projections_each=10,
    )
    dataset.gallery_overwrite = True
    dataset.build_gallery()
    assert len(os.listdir(dataset.get_gallery_dir())) == 12


def test_select_angles():
    """test the select_angles function"""
    numpy.testing.assert_allclose(
        DracRawDataset.select_angles(
            numpy.linspace(start=-20, stop=40, num=201, endpoint=True), each_angle=20
        ),
        numpy.array([-20, 0.1, 19.9, 40]),
    )

    assert DracRawDataset.select_angles((), each_angle=20) == ()
    assert DracRawDataset.select_angles((10,), each_angle=10) == (10,)
    angles = DracRawDataset.select_angles(
        numpy.linspace(start=0, stop=10, num=400, endpoint=False), each_angle=1
    )
    assert len(angles) == 11
    assert numpy.isclose(angles[0], 0, atol=0.1)
    assert numpy.isclose(angles[-1], 10, atol=0.1)



from __future__ import annotations

import os
import re
import logging
from functools import lru_cache as cache

from silx.gui import qt
from sluurp.utils import get_partitions, get_partition_gpus, get_partition_walltime

from tomwer.core.settings import SlurmSettings, SlurmSettingsMode
from tomwer.gui import icons
from tomwer.gui.utils.qt_utils import block_signals
from tomwer.gui.configuration.action import (
    BasicConfigurationAction,
    ExpertConfigurationAction,
)
from tomwer.gui.configuration.level import ConfigurationLevel

from nxtomomill.models.utils import convert_str_to_tuple

_logger = logging.getLogger(__name__)


class SlurmSettingsDialog(qt.QDialog):
    sigConfigChanged = qt.Signal()
    """Signal emit when the SlurmSetting changed"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = SlurmSettingsWindow(parent=self)
        self._mainWidget.setWindowFlags(qt.Qt.Widget)
        self.layout().addWidget(self._mainWidget)

        # buttons for validation
        self._buttons = qt.QDialogButtonBox(self)
        types = qt.QDialogButtonBox.Close
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._buttons.button(qt.QDialogButtonBox.Close).clicked.connect(self.close)

        # connect signal /slot
        self._mainWidget.sigConfigChanged.connect(self.sigConfigChanged)

    def isSlurmActive(self):
        return self._mainWidget.isSlurmActive()

    def getConfiguration(self) -> dict:
        return self._mainWidget.getConfiguration()

    def setConfiguration(self, config: dict) -> None:
        self._mainWidget.setConfiguration(config=config)


class SlurmSettingsWindow(qt.QMainWindow):
    """
    Main window to define slurm settings.
    Composed of the SlurmSettingsWidget and a combobox with some predefined settings
    """

    sigConfigChanged = qt.Signal()
    """Signal emit when the SlurmSetting changed"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)

        # define toolbar
        toolbar = qt.QToolBar(self)
        self.addToolBar(qt.Qt.TopToolBarArea, toolbar)

        self.__configurationModesAction = qt.QAction(self)
        self.__configurationModesAction.setCheckable(False)
        menu = qt.QMenu(self)
        self.__configurationModesAction.setMenu(menu)
        toolbar.addAction(self.__configurationModesAction)

        self.__configurationModesGroup = qt.QActionGroup(self)
        self.__configurationModesGroup.setExclusive(True)
        self.__configurationModesGroup.triggered.connect(self._userModeChanged)

        self._basicConfigAction = BasicConfigurationAction(toolbar)
        menu.addAction(self._basicConfigAction)
        self.__configurationModesGroup.addAction(self._basicConfigAction)
        self._expertConfiguration = ExpertConfigurationAction(toolbar)
        menu.addAction(self._expertConfiguration)
        self.__configurationModesGroup.addAction(self._expertConfiguration)

        # define maini widget
        self._mainWidget = qt.QWidget(self)
        self._mainWidget.setLayout(qt.QFormLayout())

        self._modeCombox = qt.QComboBox(self)
        self._mainWidget.layout().addRow("configuration: ", self._modeCombox)
        self._modeCombox.addItems([item.value for item in SlurmSettingsMode])
        self._modeCombox.setCurrentText(SlurmSettingsMode.GENERIC.value)

        self._settingsWidget = SlurmSettingsWidget(self, jobLimitation=None)
        self._mainWidget.layout().addRow(self._settingsWidget)

        self.setCentralWidget(self._mainWidget)

        # set up
        self._reloadPredefinedSettings()
        self._basicConfigAction.setChecked(True)
        self._userModeChanged(self._basicConfigAction)

        # connect signal / slot
        self._modeCombox.currentIndexChanged.connect(self._reloadPredefinedSettings)
        self._settingsWidget.sigConfigChanged.connect(self._configChanged)
        self._modeCombox.currentIndexChanged.connect(self.sigConfigChanged)
        # when the settings widget is edited them we automatically move to 'manual' settings. To notify visually the user
        self._settingsWidget.sigConfigChanged.connect(self._switchToManual)

    def _userModeChanged(self, action):
        self.__configurationModesAction.setIcon(action.icon())
        self.__configurationModesAction.setToolTip(action.tooltip())
        if action is self._basicConfigAction:
            level = ConfigurationLevel.OPTIONAL
        elif action is self._expertConfiguration:
            level = ConfigurationLevel.ADVANCED
        else:
            raise NotImplementedError
        self._settingsWidget.setConfigurationLevel(level)

    def _reloadPredefinedSettings(self, *args, **kkwargs):
        """
        reload settings from some predefined configuration
        """
        mode = self.getCurrentSettingsMode()
        settingsClass = SlurmSettingsMode.get_settings_class(mode)
        if settingsClass:
            with block_signals(self._settingsWidget):
                self.setConfiguration(
                    {
                        "cpu-per-task": settingsClass.N_CORES_PER_TASK,
                        "n_tasks": settingsClass.N_TASKS,
                        "n_jobs": settingsClass.N_JOBS,
                        "memory": settingsClass.MEMORY_PER_WORKER,
                        "partition": settingsClass.PARTITION,
                        "n_gpus": settingsClass.N_GPUS_PER_WORKER,
                        "job_name": settingsClass.PROJECT_NAME,
                        "walltime": settingsClass.DEFAULT_WALLTIME,
                        "python_venv": settingsClass.PYTHON_VENV,
                        "modules": settingsClass.MODULES_TO_LOAD,
                    }
                )

    def _configChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def _switchToManual(self):
        self._modeCombox.setCurrentText(SlurmSettingsMode.MANUAL.value)

    def getCurrentSettingsMode(self) -> SlurmSettingsMode:
        return SlurmSettingsMode(self._modeCombox.currentText())

    def setCurrentSettingsMode(self, mode: SlurmSettingsMode) -> SlurmSettingsMode:
        mode = SlurmSettingsMode(mode)
        self._modeCombox.setCurrentText(mode.value)

    def setMode(self, mode: SlurmSettingsMode) -> None:
        """Alias for setCurrentSettingsMode(), used by SlurmClusterOW."""
        self.setCurrentSettingsMode(mode)

    def mode(self) -> SlurmSettingsMode:
        """Alias for getCurrentSettingsMode(), used by SlurmClusterOW."""
        return self.getCurrentSettingsMode()

    # expose API
    def getConfiguration(self) -> dict:
        return self._settingsWidget.getConfiguration()

    def getSlurmClusterConfiguration(self) -> dict:
        return self._settingsWidget.getSlurmClusterConfiguration()

    def setConfiguration(self, config: dict) -> None:
        self._settingsWidget.setConfiguration(config=config)

    def isSlurmActive(self):
        return self._settingsWidget.isSlurmActive()


class SlurmSettingsWidget(qt.QWidget):
    """Widget used to define Slurm configuration to be used"""

    sigConfigChanged = qt.Signal()
    """Signal emit when the SlurmSetting changed"""

    WALL_TIME_INVALID_COLOR = qt.QColor("#FF0000")

    def __init__(
        self,
        parent=None,
        n_gpu=SlurmSettings.N_GPUS_PER_WORKER,
        jobLimitation: int | None = 1,
    ) -> None:
        """
        :param n_gpu: if provided will set the value to the number of gpu requested
                          (default is one as most of the job are nabu reconstruction)
        :param jobLimitation: if set then will hide the option to define the number of job
        """
        super().__init__(parent=parent)
        self.setLayout(qt.QGridLayout())
        """Action to display warning in case of possible 'wrong' setting"""
        # define some optional warning
        warning_icon = icons.getQIcon("warning")
        self._warning_opd_p9 = qt.QLabel("")
        self._warning_opd_p9.setPixmap(warning_icon.pixmap(20, state=qt.QIcon.On))
        self._warning_opd_p9.setToolTip(
            f"'{get_login()}' account cannot submit job on p9 architecture"
        )
        self._warning_opd_export_opts = qt.QLabel("")
        self._warning_opd_export_opts.setPixmap(
            warning_icon.pixmap(20, state=qt.QIcon.On)
        )
        self._warning_opd_export_opts.setToolTip(
            f"'{get_login()}' account needs to export job with ALL option (advanced settings)"
        )

        # queues
        self._queueLabel = qt.QLabel("queue / partition")
        self.layout().addWidget(self._queueLabel, 0, 0, 1, 1)
        self._queueCB = qt.QComboBox(self)
        self._queueCB.setEditable(True)
        for partition in get_partitions():
            if partition in ("nice*",):
                # filter some node that we cannot access in fact
                continue
            self._queueCB.addItem(partition)
        self.layout().addWidget(self._queueCB, 0, 1, 1, 1)
        self.layout().addWidget(self._warning_opd_p9, 0, 2, 1, 1)

        # n workers
        self._nWorkersLabel = qt.QLabel("number of task", self)
        self.layout().addWidget(self._nWorkersLabel, 1, 0, 1, 1)
        self._nWorkersSP = qt.QSpinBox(self)
        self._nWorkersSP.setRange(1, 100)
        self.layout().addWidget(self._nWorkersSP, 1, 1, 1, 1)

        # ncores active
        self._nCoresLabel = qt.QLabel("number of cores per task")
        self.layout().addWidget(self._nCoresLabel, 2, 0, 1, 1)
        self._nCoresSB = qt.QSpinBox(self)
        self._nCoresSB.setRange(1, 500)
        self.layout().addWidget(self._nCoresSB, 2, 1, 1, 1)

        # memory
        self._memoryLabel = qt.QLabel("memory per task")
        self.layout().addWidget(self._memoryLabel, 3, 0, 1, 1)
        self._memorySB = qt.QSpinBox(self)
        self._memorySB.setRange(1, 10000)
        self._memorySB.setSuffix("GB")
        self.layout().addWidget(self._memorySB, 3, 1, 1, 1)

        # gpu
        self._nGpuLabel = qt.QLabel("number of GPUs per task")
        self.layout().addWidget(self._nGpuLabel, 4, 0, 1, 1)
        self._nGpuSB = qt.QSpinBox(self)
        self._nGpuSB.setRange(0, 12)
        self.layout().addWidget(self._nGpuSB, 4, 1, 1, 1)

        # n jobs
        self._nJobsLabel = qt.QLabel("number of job", self)
        self.layout().addWidget(self._nJobsLabel, 5, 0, 1, 1)

        self._nJobs = qt.QSpinBox(self)
        self._nJobs.setRange(1, 100000)
        self._nJobs.setValue(1)
        self._nJobs.setToolTip("Define on how many part the job should be split in")
        self._nJobsLabel.setVisible(jobLimitation is None)
        self._nJobs.setVisible(jobLimitation is None)
        self.layout().addWidget(self._nJobs, 5, 1, 1, 1)

        # wall time
        self._wallTimeLabel = qt.QLabel("wall time", self)
        self.layout().addWidget(self._wallTimeLabel, 6, 0, 1, 1)
        self._wallTimeQLE = qt.QLineEdit("", self)
        self.__reWallTime = qt.QRegularExpression(
            r"[0-9]?[0-9]:[0-9]?[0-9]:[0-9]?[0-9]"
        )
        self._wallTimeQLE.setPlaceholderText("HH:MM:SS")
        self._wallTimeValidator = qt.QRegularExpressionValidator(
            self.__reWallTime, self
        )
        self._wallTimeQLE.setValidator(self._wallTimeValidator)
        self.layout().addWidget(self._wallTimeQLE, 6, 1, 1, 1)
        self.__defaultWallTimeBackgroundColor = self.palette().color(
            self._wallTimeQLE.backgroundRole()
        )
        self._warningWallTime = qt.QLabel("")
        self._warningWallTime.setPixmap(warning_icon.pixmap(20, state=qt.QIcon.On))
        self.layout().addWidget(self._warningWallTime, 6, 2, 1, 1)

        # python exe / modules
        self._preProcessingGroup = qt.QGroupBox("pre-processing", self)
        self._preProcessingGroup.setLayout(qt.QFormLayout())
        self._preProcessingButtonGroup = qt.QButtonGroup(self)
        self._preProcessingButtonGroup.setExclusive(True)
        self.layout().addWidget(self._preProcessingGroup, 10, 0, 1, 2)

        # python venv
        self._sourceScriptCB = qt.QRadioButton("source script (python venv)", self)
        self._pythonVenv = qt.QLineEdit("/scisoft/tomotools/activate stable", self)
        self._preProcessingButtonGroup.addButton(self._sourceScriptCB)
        self._pythonVenv.setToolTip(
            """
            Optional path to a bash script to source before executing the script.
            """
        )
        self._preProcessingGroup.layout().addRow(self._sourceScriptCB, self._pythonVenv)
        # module
        self._modulesQLE = qt.QLineEdit(", ".join(SlurmSettings.MODULES_TO_LOAD), self)
        self._modulesCB = qt.QRadioButton("module(s) to load", self)
        self._preProcessingButtonGroup.addButton(self._modulesCB)
        self._preProcessingGroup.setToolTip(
            """
            Optional list of modules to load before executing the script. each module must be separated by a coma
            """
        )
        self._preProcessingGroup.layout().addRow(self._modulesCB, self._modulesQLE)

        # job name
        self._job_nameQLabel = qt.QLabel("job name", self)
        self.layout().addWidget(self._job_nameQLabel, 13, 0, 1, 1)

        self._jobName = qt.QLineEdit("", self)
        self._jobName.setToolTip(
            """Job name. Can contains several keyword that will be formatted:
            - scan: will be replaced by the scan name
            - process: will be replaced by the process name (nabu slices, nabu volume...)
            - info: some extra information that can be provided by the process to `specify` the processing
            note: those key word have to be surrounding by surrounded `{` and `}` in order to be formatted.
            """
        )
        self.layout().addWidget(self._jobName, 13, 1, 1, 1)

        # port
        # fixme: replace by a dedicated widget / validator for TCP adress
        self._portLabel = qt.QLabel("port", self)
        self.layout().addWidget(self._portLabel, 14, 0, 1, 1)
        self._port = _PortRangeSelection(self)
        self._port.setRange(0, 99999)
        self._port.setToolTip("TCP Port for the dask-distributed scheduler")
        self.layout().addWidget(self._port, 14, 1, 1, 1)

        # dashboard port
        self._dashboardPort = qt.QSpinBox(self)
        self.layout().addWidget(self._dashboardPort, 15, 0, 1, 1)
        self._dashboardPort.setRange(0, 99999)
        self._dashboardPort.setToolTip("TCP Port for the dashboard")
        self._dashboardPortLabel = qt.QLabel("dashboard port", self)
        self.layout().addWidget(self._dashboardPort, 15, 1, 1, 1)

        # sbatch advance parameters
        self._sbatchAdvancedParameters = qt.QGroupBox("sbatch advanced settings", self)
        self._sbatchAdvancedParameters.setLayout(qt.QFormLayout())
        self.layout().addWidget(self._sbatchAdvancedParameters, 16, 0, 1, 2)

        ## export parameter
        self._exportValueCM = qt.QComboBox(self)
        self._exportValueCM.addItems(("NONE", "ALL"))
        self._exportValueCM.setItemData(
            self._exportValueCM.findText("NONE"),
            """
            Only SLURM_* variables from the user environment will be defined. User must use absolute path to the binary to be executed that will define the environment. User can not specify explicit environment variables with "NONE". However, Slurm will then implicitly attempt to load the user's environment on the node where the script is being executed, as if --get-user-env was specified. \nThis option is particularly important for jobs that are submitted on one cluster and execute on a different cluster (e.g. with different paths). To avoid steps inheriting environment export settings (e.g. "NONE") from sbatch command, the environment variable SLURM_EXPORT_ENV should be set to "ALL" in the job script.
            """,
        )
        self._exportValueCM.setItemData(
            self._exportValueCM.findText("ALL"),
            """
            All of the user's environment will be loaded (either from the caller's environment or from a clean environment if --get-user-env is specified). \nCan fail when submitting cross-platform jobs and user has some module loaded
            """,
        )
        self._exportValueCM.setCurrentText("ALL")
        self._sbatchAdvancedParameters.layout().addRow("--export", self._exportValueCM)
        self.layout().addWidget(self._warning_opd_export_opts, 16, 2, 1, 1)

        ## gpu card
        self._gpuCardCB = qt.QComboBox(self)
        self._gpuCardCB.setToolTip(
            "Specify a GPU card to be used. Using the -C command from sbatch"
        )
        self._gpuCardCB.setEditable(
            True
        )  # let the user the ability to provide a GPU that is not found for now (expecting he knows what he is doing)
        self._sbatchAdvancedParameters.layout().addRow("-C (gpu card)", self._gpuCardCB)

        # simplify gui
        self._dashboardPort.hide()
        self._dashboardPortLabel.hide()
        self._jobName.hide()
        self._job_nameQLabel.hide()
        self._portLabel.hide()  # for now we don't use the port. This can be done automatically
        self._port.hide()  # for now we don't use the port. This can be done automatically
        # for now nworker == ntask is not used
        self._nWorkersSP.setVisible(False)
        self._nWorkersLabel.setVisible(False)

        # set up the gui
        self._nCoresSB.setValue(SlurmSettings.N_CORES_PER_TASK)
        self._nWorkersSP.setValue(SlurmSettings.N_TASKS)
        self._memorySB.setValue(SlurmSettings.MEMORY_PER_WORKER)
        if self._queueCB.findText(SlurmSettings.PARTITION) >= 0:
            self._queueCB.setCurrentText(SlurmSettings.PARTITION)
        self._nGpuSB.setValue(n_gpu)
        self._jobName.setText(SlurmSettings.PROJECT_NAME)
        self._wallTimeQLE.setText(SlurmSettings.DEFAULT_WALLTIME)
        self._modulesCB.setChecked(True)  # by default we go for modules
        self._warning_opd_p9.setVisible(False)
        self._warning_opd_export_opts.setVisible(False)
        self._warningWallTime.setVisible(False)
        self._preProcessingModeChanged()
        self._partitionChanged()
        self._sbatchAdvancedSettingsChanged()
        self._nGpuChanged()

        # connect signal / slot
        self._nCoresSB.valueChanged.connect(self._configurationChanged)
        self._nWorkersSP.valueChanged.connect(self._configurationChanged)
        self._memorySB.valueChanged.connect(self._configurationChanged)
        self._queueCB.currentTextChanged.connect(self._configurationChanged)
        self._queueCB.currentTextChanged.connect(self._updateWallTimeWarnings)
        self._queueCB.currentTextChanged.connect(self._updateWallTimeTooltip)
        self._nGpuSB.valueChanged.connect(self._configurationChanged)
        self._jobName.editingFinished.connect(self._configurationChanged)
        self._wallTimeQLE.editingFinished.connect(self._configurationChanged)
        self._wallTimeQLE.editingFinished.connect(self._updateWallTimeWarnings)
        self._wallTimeQLE.editingFinished.connect(self._updateWallTimeTooltip)
        self._pythonVenv.editingFinished.connect(self._configurationChanged)
        self._modulesQLE.editingFinished.connect(self._configurationChanged)
        self._preProcessingButtonGroup.buttonClicked.connect(self._configurationChanged)
        self._port.sigRangeChanged.connect(self._configurationChanged)
        self._dashboardPort.valueChanged.connect(self._configurationChanged)
        self._preProcessingButtonGroup.buttonClicked.connect(
            self._preProcessingModeChanged
        )
        self._queueCB.currentTextChanged.connect(self._partitionChanged)
        self._nGpuSB.valueChanged.connect(self._nGpuChanged)
        self._gpuCardCB.currentTextChanged.connect(self._configurationChanged)
        self._exportValueCM.currentIndexChanged.connect(self._configurationChanged)
        self._exportValueCM.currentIndexChanged.connect(
            self._sbatchAdvancedSettingsChanged
        )

    def _updateWallTimeWarnings(self):
        time_limit, _ = self._getPartitionWallTime(partition=self.getQueue())
        palette = self._wallTimeQLE.palette()
        wall_time_limit_exceed = time_limit is not None and _WallTime(
            self.getWallTime()
        ) > _WallTime(time_limit)
        if wall_time_limit_exceed:
            palette.setColor(self.backgroundRole(), self.WALL_TIME_INVALID_COLOR)
        else:
            palette.setColor(
                self.backgroundRole(), self.__defaultWallTimeBackgroundColor
            )
        self._warningWallTime.setVisible(wall_time_limit_exceed)
        self._wallTimeQLE.setPalette(palette)

    def _updateWallTimeTooltip(self):
        partition = self.getQueue()
        time_limit, default_wall_time = self._getPartitionWallTime(partition=partition)
        if time_limit is not None and default_wall_time is not None:
            tooltip = "\n".join(
                (
                    f"for partition {partition}:",
                    f"time limit: {time_limit}",
                    f"default limit: {default_wall_time}",
                )
            )
        else:
            tooltip = f"no wall time found for {partition}"
        self._wallTimeQLE.setToolTip(tooltip)
        self._warningWallTime.setToolTip(tooltip)

    def _nGpuChanged(self, *args, **kwargs):
        nGpu = self.getNGPU()
        self._gpuCardCB.setEnabled(nGpu > 0)

    def _configurationChanged(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getNCores(self) -> int:
        return self._nCoresSB.value()

    def setNCores(self, n: int) -> None:
        self._nCoresSB.setValue(int(n))

    def getNWorkers(self) -> int:
        return self._nWorkersSP.value()

    def setNWorkers(self, n) -> None:
        self._nWorkersSP.setValue(int(n))

    def getMemory(self) -> int:
        return self._memorySB.value()

    def setMemory(self, memory: int) -> None:
        self._memorySB.setValue(int(memory))

    def getQueue(self) -> str:
        return self._queueCB.currentText()

    def setQueue(self, text: str) -> None:
        self._queueCB.setCurrentText(text)

    def getNGPU(self) -> int:
        return self._nGpuSB.value()

    def setNGPU(self, n: int) -> None:
        self._nGpuSB.setValue(int(n))

    def getNJobs(self) -> int:
        return self._nJobs.value()

    def setNJobs(self, value: int):
        self._nJobs.setValue(int(value))

    def getProjectName(self):
        return self._jobName.text()

    def setProjectName(self, name):
        self._jobName.setText(name)

    def getWallTime(self):
        parts = self._wallTimeQLE.text().split(":")
        hours = parts[0]
        if len(parts) > 1:
            minutes = parts[1]
        else:
            minutes = ""
        if len(parts) > 2:
            seconds = parts[2]
        else:
            seconds = ""

        return ":".join(
            [
                hours.zfill(2),
                minutes.zfill(2),
                seconds.zfill(2),
            ]
        )

    def setWallTime(self, walltime):
        self._wallTimeQLE.setText(walltime)
        self._updateWallTimeWarnings()
        self._updateWallTimeTooltip()

    def getPythonExe(self):
        if self._sourceScriptCB.isChecked():
            return self._pythonVenv.text()
        else:
            return None

    def setPythonExe(self, python_venv):
        self._pythonVenv.setText(python_venv)
        if python_venv != "":
            self._sourceScriptCB.setChecked(True)

    def getModulesToLoad(self) -> tuple:
        if self._modulesCB.isChecked():
            return convert_str_to_tuple(self._modulesQLE.text())
        else:
            return tuple()

    def setModulesToLoad(self, modules: tuple):
        if not isinstance(modules, (tuple, list)):
            raise TypeError(
                f"modules is expected to be a tuple or a list. Get {type(modules)} instead"
            )
        self._modulesQLE.setText(str(modules))
        if len(modules) > 0:
            self._modulesCB.setChecked(True)

    def getGpuCard(self) -> str | None:
        card = self._gpuCardCB.currentText()
        if card == "any" or self._nGpuSB == 0:
            return None
        else:
            return card

    def getSBatchExtraParams(self):
        return {
            "export": self._exportValueCM.currentText(),
            "gpu_card": self.getGpuCard(),
        }

    def setSBatchExtraParams(self, params: dict):
        export_ = params.get("export", None)
        if export_ is not None:
            index = self._exportValueCM.findText(export_)
            if index >= 0:
                self._exportValueCM.setCurrentIndex(index)
        gpu_card = params.get("gpu_card", None)
        if gpu_card is not None:
            index = self._gpuCardCB.findText("gpu_card")
            if index >= 0:
                self._gpuCardCB.setCurrentIndex(index)
            else:
                # policy when setting the extra params: if doesn't exists / found then won't set it.
                # because they can be part of .ows, this parameter is hidden by default.
                # so safer to use 'any' in the case it is unknown (debatable).
                _logger.warning(f"unable to find gpu {gpu_card}. Won't set it")

    def setConfiguration(self, config: dict) -> None:
        with block_signals(self):
            active_slurm = config.get("active_slurm", None)
            if active_slurm is not None:
                self._slurmCB.setChecked(active_slurm)

            n_cores = config.get("cpu-per-task", None)
            if n_cores is not None:
                self.setNCores(n_cores)

            n_workers = config.get("n_tasks", None)
            if n_workers is not None:
                self.setNWorkers(n_workers)

            memory = config.get("memory", None)
            if memory is not None:
                if isinstance(memory, str):
                    memory = memory.replace(" ", "").lower().rstrip("gb")
                self.setMemory(int(memory))

            queue_ = config.get("partition", None)
            if queue_ is not None:
                queue_ = queue_.rstrip("'").rstrip('"')
                queue_ = queue_.lstrip("'").lstrip('"')
                self.setQueue(queue_)

            n_gpu = config.get("n_gpus", None)
            if n_gpu is not None:
                self.setNGPU(int(n_gpu))

            project_name = config.get("job_name", None)
            if project_name is not None:
                self.setProjectName(project_name)

            wall_time = config.get("walltime", None)
            if wall_time is not None:
                self.setWallTime(wall_time)

            python_venv = config.get("python_venv", None)
            if python_venv is not None:
                python_venv = python_venv.rstrip("'").rstrip('"')
                python_venv = python_venv.lstrip("'").lstrip('"')
                self.setPythonExe(python_venv)

            modules = config.get("modules", None)
            if modules is not None:
                modules = convert_str_to_tuple(modules)
                self.setModulesToLoad(modules)

            sbatch_extra_params = config.get("sbatch_extra_params", {})
            self.setSBatchExtraParams(sbatch_extra_params)

            n_jobs = config.get("n_jobs", None)
            if n_jobs is not None:
                self.setNJobs(n_jobs)
            self._preProcessingModeChanged()  # make sure modules and python venv is enabled according to the activate mode

        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        config = {
            "cpu-per-task": self.getNCores(),
            "n_tasks": self.getNWorkers(),
            "n_jobs": self.getNJobs(),
            "memory": self.getMemory(),
            "partition": self.getQueue(),
            "n_gpus": self.getNGPU(),
            "job_name": self.getProjectName(),
            "walltime": self.getWallTime(),
            "sbatch_extra_params": self.getSBatchExtraParams(),
        }
        if self._modulesCB.isChecked():
            config["modules"] = self.getModulesToLoad()
        elif self._sourceScriptCB.isChecked():
            config["python_venv"] = self.getPythonExe()
        else:
            raise ValueError("'modules' or python environment should be enable")
        return config

    def getSlurmClusterConfiguration(self):
        from tomwer.core.cluster import SlurmClusterConfiguration

        return SlurmClusterConfiguration().from_dict(self.getConfiguration())

    def _preProcessingModeChanged(self):
        self._modulesQLE.setEnabled(self._modulesCB.isChecked())
        self._pythonVenv.setEnabled(self._sourceScriptCB.isChecked())

    def setConfigurationLevel(self, level: ConfigurationLevel):
        self._sbatchAdvancedParameters.setVisible(level >= ConfigurationLevel.ADVANCED)
        self._wallTimeLabel.setVisible(level >= ConfigurationLevel.ADVANCED)
        self._wallTimeQLE.setVisible(level >= ConfigurationLevel.ADVANCED)

    def _partitionChanged(self, *args, **kwargs):
        partition = self.getQueue()
        # op(erator) accounts (from icc computers) cannot export to p9. they don't have home directories
        self._warning_opd_p9.setVisible(
            current_login_is_op_account() and partition.lower().startswith("p9")
        )
        gpus = self._getGpus(partition=partition)
        self._gpuCardCB.clear()
        self._gpuCardCB.addItems(gpus)
        self._gpuCardCB.setCurrentText("any")

    def _sbatchAdvancedSettingsChanged(self, *args, **kwargs):
        # op(erator) accounts (from icc computers) must export ALL environment variable. Else job fails
        # warning: we want to display the warning no matter if we are in 'advance options' or not.
        # this way the user always see the warning when necessary
        self._warning_opd_export_opts.setVisible(
            current_login_is_op_account()
            and self._exportValueCM.currentText() == "NONE"
        )

    @cache(maxsize=None)
    def _getGpus(self, partition) -> tuple:
        try:
            gpus = get_partition_gpus(partition)
        except Exception as e:
            _logger.error(f"Failed to detect GPU on {partition}. Error is {e}")
            gpus = ("any",)
        else:
            gpus = list(gpus) + [
                "any",
            ]
            return gpus

    @cache(maxsize=None)
    def _getPartitionWallTime(self, partition: str) -> tuple[str | None, str | None]:
        try:
            times = get_partition_walltime(partition=partition)
        except Exception:
            return None, None
        else:
            return times["time"], times["default_time"]


class _PortRangeSelection(qt.QWidget):
    sigRangeChanged = qt.Signal()
    """Signal emit when the port range change"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)
        self.setLayout(qt.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        # from label
        self._fromLabel = qt.QLabel("from", self)
        self._fromLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._fromLabel)
        self._fromQSpinBox = qt.QSpinBox(self)
        self.layout().addWidget(self._fromQSpinBox)
        # to label
        self._toLabel = qt.QLabel("to", self)
        self._toLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._toLabel)
        self._toQSpinBox = qt.QSpinBox(self)
        self.layout().addWidget(self._toQSpinBox)
        # steps label
        self._stepLabel = qt.QLabel("step", self)
        self._stepLabel.setAlignment(qt.Qt.AlignRight | qt.Qt.AlignVCenter)
        self.layout().addWidget(self._stepLabel)
        self._stepQSpinBox = qt.QSpinBox(self)
        self.layout().addWidget(self._stepQSpinBox)

        # connect signal / slot
        self._fromQSpinBox.valueChanged.connect(self._rangeChanged)
        self._toQSpinBox.valueChanged.connect(self._rangeChanged)
        self._stepQSpinBox.valueChanged.connect(self._rangeChanged)

    def _rangeChanged(self, *args, **kwargs):
        self.sigRangeChanged.emit()

    def getRange(self) -> tuple:
        return (
            self._fromQSpinBox.value(),
            self._toQSpinBox.value(),
            self._stepQSpinBox.value(),
        )

    def setRange(self, min_: int, max_: int, step: int | None = None) -> None:
        self._fromQSpinBox.setValue(min(min_, max_))
        self._toQSpinBox.setValue(max(min_, max_))
        if step is not None:
            self._stepQSpinBox.setValue(step)


def get_login() -> str | None:
    try:
        return os.getlogin()
    except OSError:
        """In CI computer this raises the following error
        OSError: [Errno 6] No such device or address
        """
        return None


def current_login_is_op_account() -> bool:
    """
    return True if the current login is a op[i]dXX login.
    """
    return is_op_account(get_login())


def is_op_account(user: str | None) -> bool:
    """
    Determines if a given username follows the 'op' account format,
    specifically 'opdXX' or 'opidXX', where 'XX' are numeric digits.

    Args:
        user (Optional[str]): The username to check.

    Returns:
        bool: True if the username matches the pattern, False otherwise.
    """

    if user is None:
        return False

    op_account_pattern = r"^opi?d\d{2}$"
    return bool(re.match(op_account_pattern, user))


class _WallTime:
    """Util class to compare two walltime provided as D-HH:MM:SS or as HH:MM:SS"""

    def __init__(self, walltime: str):
        parts = walltime.split("-")
        if len(parts) > 1:
            assert len(parts) == 2, f"unexpected walltime format ({walltime})"
            days = int(parts[0])
        else:
            days = 0

        def minutes_to_seconds(value):
            return value * 60

        def hours_to_seconds(value):
            return minutes_to_seconds(value * 60)

        def days_to_seconds(value):
            return hours_to_seconds(value * 24)

        hours, minutes, seconds = map(int, parts[-1].split(":"))
        self.value = (
            seconds
            + minutes_to_seconds(minutes)
            + hours_to_seconds(hours)
            + days_to_seconds(days)
        )

    def __lt__(self, other):
        return self.value < other.value

    def __gt__(self, other):
        return self.value > other.value

    def __eq__(self, other):
        return self.value == other.value



# coding: utf-8

"""
This module is dedicated to cast of volume from one file format to the other.
"""
from __future__ import annotations


import logging
import os

import numpy
import pathlib

from nabu.io.cast_volume import RESCALE_MAX_PERCENTILE, RESCALE_MIN_PERCENTILE
from nabu.io.cast_volume import cast_volume as _nabu_cast_volume
from nabu.io.cast_volume import get_default_output_volume
from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from sluurp.executor import submit as submit_to_slurm_cluster
from sluurp.job import SBatchScriptJob
from tomoscan.volumebase import VolumeBase
from tomoscan.utils.io import filter_esrf_mounting_points

from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.reconstruction.nabu import settings
from tomwer.core.process.output import NabuOutputFileFormat
from tomwer.core.process.reconstruction.nabu.utils import (
    get_default_output_volume_for_tiff_3d,
)
from tomwer.core.process.task import Task
from tomwer.core.utils.scanutils import (
    format_output_location as format_scan_output_location,
)
from tomwer.core.utils.slurm import get_slurm_script_name
from tomwer.core.utils.volumeutils import volume_identifier_to_volume
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.core.process.drac.processeddataset import (
    DracReconstructedVolumeDataset,
)

_logger = logging.getLogger(__name__)

CAST_VOLUME_WORKING_SUB_DIRECTORY = "cast_volume"

DEFAULT_OUTPUT_DIR = (
    "{volume_data_parent_folder}" f"/{CAST_VOLUME_WORKING_SUB_DIRECTORY}"
)


class CastVolumeTask(
    Task,
    SuperviseProcess,
    input_names=(
        "volume",
        "configuration",
    ),
    optional_input_names=("scan", "output_volume"),
    output_names=(
        "volume",
        "future_tomo_obj",
        "data_portal_processed_datasets",
    ),
):
    def __init__(
        self,
        process_id=None,
        varinfo=None,
        inputs=None,
        node_id=None,
        node_attrs=None,
        execinfo=None,
    ):
        SuperviseProcess.__init__(self, process_id=process_id)
        Task.__init__(
            self,
            varinfo=varinfo,
            inputs=inputs,
            node_id=node_id,
            node_attrs=node_attrs,
            execinfo=execinfo,
        )
        self._cluster_config = None

    def run(self):
        input_volume = volume_identifier_to_volume(self.inputs.volume)
        if not isinstance(input_volume, VolumeBase):
            raise TypeError(
                f"input_volume is a {type(input_volume)} when {VolumeBase} expected"
            )
        configuration = self.inputs.configuration
        self._cluster_config = configuration.pop("cluster_config", None)
        if isinstance(self._cluster_config, SlurmClusterConfiguration):
            self._cluster_config = self._cluster_config.to_dict()

        if self.inputs.scan:
            scan = self.inputs.scan
        else:
            scan = None

        if self.inputs.output_volume:
            output_volume = self.inputs.output_volume
            if not isinstance(output_volume, TomwerVolumeBase):
                output_volume = VolumeFactory.create_tomo_object_from_identifier(
                    output_volume
                )
        else:
            output_file_format = configuration.get(
                "output_file_format", None
            )  # expects values like hdf5, edf ...
            output_dir = configuration.get("output_dir", None)
            if output_file_format is None or output_dir is None:
                mess = "'output_volume' or ('output_file_format' and 'output_dir' from configuration) should be provided."
                _logger.processFailed(f"volume cast of {str(input_volume)} failed")
                ProcessManager().notify_dataset_state(
                    dataset=scan or input_volume,
                    process=self,
                    state=DatasetState.FAILED,
                )
                raise ValueError(mess)

            try:
                output_dir = TomwerVolumeBase.format_output_location(
                    location=output_dir,
                    volume=input_volume,
                )
            except KeyError:
                pass

            if scan:
                try:
                    output_dir = format_scan_output_location(
                        location=output_dir,
                        scan=scan,
                    )
                except KeyError:
                    pass
            output_file_format = NabuOutputFileFormat.from_value(output_file_format)
            if output_file_format is NabuOutputFileFormat.TIFF_3D:
                output_volume = get_default_output_volume_for_tiff_3d(
                    input_volume=input_volume,
                    output_dir=output_dir,
                )
            else:
                output_volume = get_default_output_volume(
                    input_volume=input_volume,
                    output_type=output_file_format.value,
                    output_dir=output_dir,
                )
            # convert from tomoscan volume to tomwer volume
            output_volume = VolumeFactory.create_tomo_object_from_identifier(
                output_volume.get_identifier().to_str()
            )

        # build output volume object
        overwrite = configuration.get("overwrite", False)
        remove_input_volume = configuration.get("remove_input_volume", False)
        output_volume.overwrite = overwrite
        cratios = configuration.get("compression_ratios", None)
        if cratios is not None:
            output_volume.cratios = cratios

        output_data_type = numpy.dtype(
            configuration.get("output_data_type")
        )  # expects values like numpy uint8, unint16...

        rescale_min_percentile = configuration.get(
            "rescale_min_percentile", RESCALE_MIN_PERCENTILE
        )
        rescale_max_percentile = configuration.get(
            "rescale_max_percentile", RESCALE_MAX_PERCENTILE
        )
        data_min = configuration.get("data_min", None)
        data_max = configuration.get("data_max", None)

        # run volume cast locally
        if self._cluster_config is None:
            try:
                _nabu_cast_volume(
                    input_volume=input_volume,
                    output_volume=output_volume,
                    output_data_type=output_data_type,
                    data_min=data_min,
                    data_max=data_max,
                    scan=scan,
                    rescale_min_percentile=rescale_min_percentile,
                    rescale_max_percentile=rescale_max_percentile,
                    save=True,
                    store=False,
                    remove_input_volume=remove_input_volume,
                )
            except Exception as e:
                mess = f"volume cast of {str(input_volume)} failed. Reason is {str(e)}"
                _logger.processFailed(mess)
                state = DatasetState.FAILED
            else:
                mess = f"volume cast of {str(input_volume)} succeed"
                _logger.processSucceed(mess)
                state = DatasetState.SUCCEED

            ProcessManager().notify_dataset_state(
                dataset=scan or input_volume,
                process=self,
                state=state,
                details=mess,
            )
            self.outputs.future_tomo_obj = None
            self.outputs.volume = output_volume
            if scan is not None:
                scan.cast_volume = output_volume.get_identifier()
            else:
                input_volume.cast_volume = output_volume.get_identifier()
            self.outputs.data_portal_processed_datasets = tuple()
        # run volume cast remotely
        else:

            def get_command():
                command = f"python3 -m {settings.NABU_CAST_APP_PATH} '{input_volume.get_identifier().to_str()}'"
                command += (
                    f" --output_volume='{output_volume.get_identifier().to_str()}'"
                )
                if overwrite:
                    command += " --overwrite"
                command += f" --output-data-type={str(output_data_type)}"
                if data_min is not None:
                    command += f" --data_min={data_min}"
                if data_max is not None:
                    command += f" --data_max={data_max}"
                if rescale_min_percentile is not None:
                    command += f" --rescale_min_percentile={rescale_min_percentile}"
                if rescale_max_percentile is not None:
                    command += f" --rescale_max_percentile={rescale_max_percentile}"
                if remove_input_volume:
                    command += " --remove-input-volume"
                return command

            script_name = get_slurm_script_name(prefix="nabu_cast")
            if scan is not None:
                working_directory = scan.working_directory.resolve()
                script_path = os.path.join(
                    scan.resolved_path, "slurm_scripts", script_name
                )
            else:
                working_directory = os.path.dirname(input_volume.data_url.file_path())
                working_directory = str(pathlib.Path(working_directory).resolve())
                script_path = os.path.join(
                    working_directory, "slurm_scripts", script_name
                )
            # for now force job name
            self._cluster_config["job_name"] = (
                f"tomwer-cast-volume {input_volume.get_identifier().to_str()} to {output_volume.get_identifier().to_str()}"
            )
            job = SBatchScriptJob(
                slurm_config=self._cluster_config,
                script=(get_command(),),
                script_path=filter_esrf_mounting_points(script_path),
                clean_script=False,
                working_directory=filter_esrf_mounting_points(working_directory),
            )
            future_slurm_job = submit_to_slurm_cluster(job)
            if scan is not None:
                future_obj = scan
            else:
                future_obj = output_volume
            self.outputs.future_tomo_obj = FutureTomwerObject(
                tomo_obj=future_obj,
                futures=(future_slurm_job,),
                process_requester_id=self.process_id,
            )
            self.outputs.volume = output_volume

            if scan is not None:
                drac_reconstructed_dataset = DracReconstructedVolumeDataset(
                    tomo_obj=output_volume,
                    source_scan=scan,
                )
                self.outputs.data_portal_processed_datasets = (
                    drac_reconstructed_dataset,
                )
            # for now at the task level we consider this is succeed if it has been submitted
            ProcessManager().notify_dataset_state(
                dataset=scan or input_volume,
                process=self,
                state=DatasetState.SUCCEED,
                details="cast has been correctly submitted to slurm",
            )



# coding: utf-8
from __future__ import annotations

import os
import shutil
import tempfile
import unittest

from tomwer.core.process.script.python import PythonScript
from tomwer.core.utils.scanutils import MockNXtomo


class TestPythonScript(unittest.TestCase):
    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        dim = 10
        mock = MockNXtomo(
            scan_path=os.path.join(self.tempdir, "scan1"),
            n_proj=10,
            n_ini_proj=10,
            scan_range=180,
            dim=dim,
        )
        self.scan = mock.scan

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def test(self):
        process = PythonScript(inputs={"data": self.scan})
        process.definition()
        process.program_version()
        process.program_name()
        # TODO: configuration should be passed in inputs during construction
        process.set_configuration(
            {
                "scriptText": "print('toto')",
            }
        )

        process.run()
